
> module Context where
>
> import Prelude hiding (pi)
> import Abstract
> import Parsing (isWhite)
> import List
>
> data Context                 = Context { pb     :: !Bool
>                                        , pem    :: !Bool
>                                        , pi     :: !Bool
>                                        , ps     :: !Bool
>                                        , ptt    :: !Bool
>                                        , pu     :: !Int
>                                        , psize  :: !Int
>                                        , pcolor :: !Color
>                                        }
>                              deriving Eq
>
> type DecoChar                = (Char,Context)
>
> data Walk a                  = CW { context :: !Context
>                                   , prev    :: Maybe DecoChar
>                                   , tree    :: a
>                                   }
>
> contextWalk                  :: Walk SMLNG -> Walk SMLNG
> contextWalk cw               =  case tree cw of
>   Text s                     -> let CW c  p  _  = cw
>                                     CW c' p' s' = filterSpaces (CW c p s)
>                                 in  CW c' p' (Text s')
>   Tag ts bds                 -> let CW c  p  _  = cw
>                                     c_int       = modifyContext c ts
>                                     CW c' p' s' = walkList (CW c_int p bds)
>                                     -- join tags
>                                     (ts',bds')  = case s' of 
>                                      [Tag ts_ bds_] -> 
>                                             (ts ++ ts_,bds_)
>                                      _              -> (ts,s')
>                                 in  case bds' of
>                                      [] -> CW c p' (Text "")
>                                      [Text ""] -> CW c p' (Text "")
>                                      _ -> CW c  p' (Tag (simplifyTags c ts') bds')

> simplifyTags                 :: Context -> [Tag] -> [Tag]
> -- we can safely eliminate all but the last color tag,
> -- all but the last size tag, any double tags (except EM,U), an EM tag if
> -- there is also a S tag, all tags that do not modify the context
> -- we do all this by comparing the contexts; after simplification,
> -- the tags are commutative (with exception of PL).
> simplifyTags c ts            =  contextDiff c (modifyContext c ts) []
>   where contextDiff c1 c2    =  (if   (pb c1 && not (pb c2))
>                                    || (pi c1 && not (pi c2))
>                                    || (ps c1 && not (ps c2))
>                                    || (ptt c1 && not (ptt c2))
>                                    || (pu c1 > pu c2)
>                                  then (TPL:) else id)
>                                .(if   not (pb c1) && pb c2
>                                  then (TB:) else id)
>                                .(if   (pem c1 /= pem c2) && not (ps c2)
>                                  then (TEM:) else id)
>                                .(if   not (pi c1) && pi c2
>                                  then (TI:) else id)
>                                .(if   not (ps c1) && ps c2
>                                  then (TS:) else id)
>                                .(if   not (ptt c1) && ptt c2
>                                  then (TTT:) else id)
>                                .(if   pu c1 < pu c2
>                                  then foldr (.) id (replicate (pu c2-pu c1)
>                                                               (TU:))
>                                  else id)
>                                .(if   psize c1 /= psize c2
>                                  then ((Tsize (psize c2)):) else id)
>                                .(if   pcolor c1 /= pcolor c2
>                                  then ((Tcolor (pcolor c2)):) else id)

> modifyContext                :: Context -> [Tag] -> Context
> modifyContext                =  foldl mc
>   where mc c TB              =  c { pb = True }
>         mc c TEM             =  c { pem = if ps c then pem c else not (pem c) }
>         mc c TI              =  c { pi = True }
>         mc c TPL             =  c { pu = 0, pb = False, pem = False
>                                   , pi = False, ps = False, ptt = False }
>         mc c TS              =  c { ps = True }
>         mc c TTT             =  c { ptt = True }
>         mc c TU              =  c { pu = if (pu c<3) then pu c+1 else pu c } 
>         mc c (Tsize n)       =  c { psize = n }
>         mc c (Tcolor n)      =  c { pcolor = n }

> walkList                     :: Walk [SMLNG] -> Walk [SMLNG]
> walkList cw                  =  case tree cw of
>   []                         -> cw
>   (x:xs)                     -> let CW c   p   _  = cw
>                                     CW c'  p'  s' = contextWalk (CW c p x)
>                                     CW c'' p'' s''= walkList (CW c' p' xs)
>                                     s_ = case s' of Text "" -> id
>                                                     _       -> (s':)
>                                 in  CW c'' p'' (s_ s'')
>
> filterSpaces                 :: Walk String -> Walk String
> filterSpaces cw              =  case tree cw of
>   []                         -> cw
>   (x:xs)                     -> let (t,p) =
>                                       if   isWhite x && alsoWhite (prev cw) 
>                                         && spaceeq (context cw) (prev cw)
>                                       then ([] ,prev cw) -- filter the space
>                                       else ([x],Just (x,context cw)) 
>                                                          -- not filter
>                                     cw'   = filterSpaces cw { prev = p 
>                                                             , tree = xs
>                                                             }
>                                 in  cw' { tree = t ++ tree cw' }

> alsoWhite                    :: Maybe DecoChar -> Bool
> alsoWhite Nothing            =  False
> alsoWhite (Just (c,_))       =  isWhite c

> spaceeq                      :: Context -> Maybe DecoChar -> Bool
> spaceeq _ Nothing            =  False
> spaceeq c1 (Just (_,c2)) 
>   | ptt c1                   =  False
>   | otherwise                =  psize c1 == psize c2
>                              && pu c1 == pu c2
>                              && puc c1 == puc c2
>  where puc  c | pu c == 0    =  Cw
>               | otherwise    =  pcolor c

> dowalk                       :: SMLNG -> SMLNG
> dowalk t                     =  tree $ contextWalk (CW initContext Nothing t)
>   where initContext          = Context { pb  = False
>                                        , pem = False
>                                        , pi  = False
>                                        , ps  = False
>                                        , ptt = False
>                                        , pu  = 0
>                                        , psize = -1
>                                        , pcolor = Cdef
>                                        }